==> gonzui-0.4.tar.gz: gonzui-0.4/gonzui/texttokenizer.c <==
/* -*- mode: C; indent-tabs-mode: nil; c-basic-offset: 2 c-style: "BSD" -*- */
/* 
 *
 * TextTokenizer - a simple text tokenizer
 *
 * Copyright (C) 2005 Satoru Takabayashi <satoru@namazu.org> 
 * Copyright (C) 2005 Keisuke Nishida <knishida@open-cobol.org>
 *     All rights reserved.
 *     This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * You can redistribute it and/or modify it under the terms of 
 * the GNU General Public License version 2.
 *
 *
 */

#include <assert.h>
#include <ruby.h>
#include <st.h>

typedef struct gonzui_texttokenizer {
  unsigned char *text;
  unsigned char *eot;
  int len;
} gonzui_texttokenizer_t;

static gonzui_texttokenizer_t *texttokenizer_new(VALUE text)
{
  struct RString *s = RSTRING(rb_obj_as_string(text));
  gonzui_texttokenizer_t *p = malloc(sizeof(gonzui_texttokenizer_t));
  memset(p, 0, sizeof(gonzui_texttokenizer_t));
  p->text = s->ptr;
  p->len = s->len;
  p->eot = p->text + p->len;
  return p;
}

static void texttokenizer_free(gonzui_texttokenizer_t *p)
{
  if (p == NULL)
    return;
  free(p);
}

static VALUE texttokenizer_s_allocate(VALUE klass)
{
  return Data_Wrap_Struct(klass, NULL, texttokenizer_free, NULL);
}

static VALUE texttokenizer_initialize(VALUE self, VALUE text)
{
  gonzui_texttokenizer_t *p;
  Data_Get_Struct(self, gonzui_texttokenizer_t, p);
  if (p)
    rb_raise(rb_eArgError, "called twice");

  DATA_PTR(self) = texttokenizer_new(text);
  return self;
}


static inline int utf8len(const unsigned char *s, const unsigned char *eot)
{
  int len = 0;
  if (*s < 0x80) {
    len = 1;
  } else if ((s + 1 < eot) && (*s & 0xe0) == 0xc0) {
    len = 2;
  } else if ((s + 2 < eot) && (*s & 0xf0) == 0xe0) {
    len = 3;
  } else if ((s + 3 < eot) && (*s & 0xf8) == 0xf0) {
    len = 4;
  } else if ((s + 4 < eot) && (*s & 0xfc) == 0xf8) {
    len = 5;
  } else if ((s + 5 < eot) && (*s & 0xfe) == 0xfc) {
    len = 6;
  } else {
    /* invalid character */
    rb_raise(rb_eArgError, "invalid UTF-8 character");
  }
  return len;
}

static inline unsigned char *
skip(unsigned char *s, unsigned char *eot, int *lineno)
{
  for (; s < eot; s++) {
    if (*s == '\n') {
      (*lineno)++;
    } else if (isalnum(*s) || *s >= 0x80) {
      break;
    }
  }
  return s;
}

/*
 * Iterate over each word.
 * word:  [a-zA-Z0-9]+ or single multi-byte UTF-8 character
 */
static VALUE texttokenizer_each(VALUE self)
{
  int lineno = 0;
  unsigned char *s;
  gonzui_texttokenizer_t *p;
  Data_Get_Struct(self, gonzui_texttokenizer_t, p);
  if (p == NULL)
    assert(0);

  s = p->text;
  s = skip(s, p->eot, &lineno);

  while (s < p->eot) {
    unsigned char *b = s;
    if (*s >= 0x80) {
      s += utf8len(s, p->eot);
    } else {
      for (; s < p->eot; s++)
        if (!((isalnum(*s) || *s == '_')))
          break;
    }
    rb_yield_values(3, rb_str_new(b, s - b),
                    INT2FIX(b - p->text), INT2FIX(lineno));
    s = skip(s, p->eot, &lineno);
  }
  return Qnil;
}

void Init_texttokenizer()
{
  VALUE Gonzui = rb_define_module("Gonzui");
  VALUE TextTokenizer = rb_define_class_under(Gonzui, "TextTokenizer", 
                                              rb_cData);
  rb_define_alloc_func(TextTokenizer, texttokenizer_s_allocate);
  rb_define_method(TextTokenizer, "initialize", texttokenizer_initialize, 1);
  rb_define_method(TextTokenizer, "each", texttokenizer_each, 0);
}

==> gonzui-1.1.tar.gz: gonzui-1.1/gonzui/texttokenizer.c <==
/* -*- mode: C; indent-tabs-mode: nil; c-basic-offset: 2 c-style: "BSD" -*- */
/* 
 * texttokenizer.c - a simple text tokenizer
 *
 * Copyright (C) 2005 Satoru Takabayashi <satoru@namazu.org> 
 * Copyright (C) 2005 Keisuke Nishida <knishida@open-cobol.org>
 *     All rights reserved.
 *     This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * You can redistribute it and/or modify it under the terms of 
 * the GNU General Public License version 2.
 *
 *
 */

#include <assert.h>
#include <ruby.h>

static inline int utf8len(const unsigned char *s, const unsigned char *eot)
{
  int len = 0;
  if (*s < 0x80) {
    len = 1;
  } else if ((s + 1 < eot) && (*s & 0xe0) == 0xc0) {
    len = 2;
  } else if ((s + 2 < eot) && (*s & 0xf0) == 0xe0) {
    len = 3;
  } else if ((s + 3 < eot) && (*s & 0xf8) == 0xf0) {
    len = 4;
  } else if ((s + 4 < eot) && (*s & 0xfc) == 0xf8) {
    len = 5;
  } else if ((s + 5 < eot) && (*s & 0xfe) == 0xfc) {
    len = 6;
  } else {
    rb_raise(rb_eArgError, "invalid UTF-8 character");
  }
  return len;
}

static inline unsigned char *
skip(unsigned char *s, unsigned char *eot)
{
  for (; s < eot; s++)
    if (isalnum(*s) || *s >= 0x80)
      break;
  return s;
}

/*
 * Iterate over each word.
 * word:  [a-zA-Z0-9]+ or single multi-byte UTF-8 character
 */
static VALUE texttokenizer_each_word(VALUE obj, VALUE text)
{
  struct RString *str;
  unsigned char *s, *beg, *eot;

  str = RSTRING(rb_obj_as_string(text));
  beg = str->ptr;
  eot = beg + str->len;
  s = skip(beg, eot);

  while (s < eot) {
    unsigned char *b = s;
    if (*s >= 0x80) {
      s += utf8len(s, eot);
    } else {
      for (; s < eot; s++)
        if (!((isalnum(*s) || *s == '_')))
          break;
    }
    rb_yield_values(2, rb_str_new(b, s - b), INT2FIX(b - beg));
    s = skip(s, eot);
  }
  return Qnil;
}

void Init_texttokenizer()
{
  VALUE mGonzui, mTextTokenizer;
  mGonzui = rb_define_module("Gonzui");
  mTextTokenizer = rb_define_module_under(mGonzui, "TextTokenizer");

  rb_define_module_function(mTextTokenizer, "each_word",
                            texttokenizer_each_word, 1);
}
